#include <Adafruit_Arcada.h>

#if defined(SD_CS) && defined(USE_SD_FS)
  static SdFat FileSys(&SD_SPI_PORT);
#elif defined(USE_QSPI_FS)
  static Adafruit_SPIFlash flash(PIN_QSPI_SCK, PIN_QSPI_IO1, PIN_QSPI_IO0, PIN_QSPI_CS);
  static Adafruit_M0_Express_CircuitPython FileSys(flash);
#endif

Adafruit_ST7735 tft = Adafruit_ST7735(&TFT_SPI, TFT_CS,  TFT_DC, TFT_RST);

Adafruit_Arcada::Adafruit_Arcada(void) {
}

/**************************************************************************/
/*!
    @brief  Initialize GPIO, display, NeoPixels, TFT, sound system, etc.
    @return True on success, False if something failed!
*/
/**************************************************************************/
bool Adafruit_Arcada::begin(void) {
  setBacklight(0);

  // current working dir is /
  strcpy(_cwd_path, "/");

  pixels.updateLength(NEOPIXEL_NUM);
  pixels.setPin(NEOPIXEL_PIN);
  pixels.begin();
  pixels.setBrightness(20);
  pixels.show();  // turn off

#ifdef ARCADA_BUTTONPIN_START
  pinMode(ARCADA_BUTTONPIN_START, INPUT_PULLUP);
#endif

#ifdef ARCADA_BUTTONPIN_SELECT
  pinMode(ARCADA_BUTTONPIN_SELECT, INPUT_PULLUP);
#endif

#ifdef ARCADA_BUTTONPIN_A
  pinMode(ARCADA_BUTTONPIN_A, INPUT_PULLUP);
#endif

#ifdef ARCADA_BUTTONPIN_B
  pinMode(ARCADA_BUTTONPIN_B, INPUT_PULLUP);
#endif

#ifdef ARCADA_BUTTONPIN_UP  // gpio for buttons
  pinMode(ARCADA_BUTTONPIN_UP, INPUT_PULLUP);
  pinMode(ARCADA_BUTTONPIN_DOWN, INPUT_PULLUP);
  pinMode(ARCADA_BUTTONPIN_LEFT, INPUT_PULLUP);
  pinMode(ARCADA_BUTTONPIN_RIGHT, INPUT_PULLUP);
#endif

#ifdef BUTTON_CLOCK
  pinMode(BUTTON_CLOCK, OUTPUT);
  digitalWrite(BUTTON_CLOCK, HIGH);
  pinMode(BUTTON_LATCH, OUTPUT);
  digitalWrite(BUTTON_LATCH, HIGH);
  pinMode(BUTTON_DATA, INPUT);
#endif

  
  TFT_INIT;
  tft.fillScreen(TFT_DEFAULTFILL);
  tft.setRotation(TFT_ROTATION);
  setBacklight(255);
  _first_frame = true;
  _framebuffer = NULL;

  return true;
}

void Adafruit_Arcada::setBacklight(uint8_t brightness) {
  pinMode(TFT_LITE, OUTPUT);
  analogWrite(TFT_LITE, brightness);
}

void Adafruit_Arcada::enableSpeaker(bool on) {
  pinMode(SPEAKER_ENABLE, OUTPUT);
  digitalWrite(SPEAKER_ENABLE, on);
}

/**************************************************************************/
/*!
    @brief printf wrapper to serial debug, handy for logging config, C inclusion
    @param format The printf-compatible format and extra args
*/
/**************************************************************************/
void Adafruit_Arcada::printf(const char *format, ...) {
  va_list args;
  va_start(args, format);
  
  Serial.printf(format, args);
  
  va_end(args);
}


void Adafruit_Arcada::print(const char *s) {
   Serial.print(s);
}

void Adafruit_Arcada::println(const char *s) {
   Serial.println(s);
}


void Adafruit_Arcada::print(int32_t d, uint8_t format) {
  Serial.print(d, format);
}

void Adafruit_Arcada::println(int32_t d, uint8_t format) {
  Serial.println(d, format);
}

/**************************************************************************/
/*!
    @brief  Read X analog joystick
    @return Signed 16 bits, from -512 to 511, 0 being 'center'
*/
/**************************************************************************/
int16_t Adafruit_Arcada::readJoystickX(uint8_t sampling) {

  float reading = 0;
#ifdef ARCADA_JOYSTICK_X
  for (int i=0; i<sampling; i++) {
    reading += analogRead(ARCADA_JOYSTICK_X);
  }
  reading /= sampling;

  // adjust range from 0->1024 to -512 to 511;
  reading -= _joyx_center;
#endif
  return reading;
}

/**************************************************************************/
/*!
    @brief  Read Y analog joystick
    @return Signed 16 bits, from -512 to 511, 0 being 'center'
*/
/**************************************************************************/
int16_t Adafruit_Arcada::readJoystickY(uint8_t sampling) {

  float reading = 0;
#ifdef ARCADA_JOYSTICK_Y
  for (int i=0; i<sampling; i++) {
    reading += analogRead(ARCADA_JOYSTICK_Y);
  }
  reading /= sampling;

  // adjust range from 0->1024 to -512 to 511;
  reading -= _joyy_center;
#endif
  return reading;
}

/**************************************************************************/
/*!
    @brief  Read all buttons/joystick and return a bitmask of which buttons are
            pressed, check ARCADA_BUTTONMASK_* for valid bitmasks to check again.
            If there's an analog joystick, it will 'simulate' the button presses.
    @return Bit array with up to 32 buttons, 1 for pressed, 0 for not.
*/
/**************************************************************************/
uint32_t Adafruit_Arcada::readButtons(void) {
  uint32_t buttons = 0;


#ifdef BUTTON_CLOCK
  // Use a latch to read 8 bits
  uint8_t shift_buttons = 0;
  digitalWrite(BUTTON_LATCH, LOW);
  delayMicroseconds(1);
  digitalWrite(BUTTON_LATCH, HIGH);
  delayMicroseconds(1);
  
  for(int i = 0; i < 8; i++) {
    shift_buttons <<= 1;
    shift_buttons |= digitalRead(BUTTON_DATA);
    digitalWrite(BUTTON_CLOCK, HIGH);
    delayMicroseconds(1);
    digitalWrite(BUTTON_CLOCK, LOW);
    delayMicroseconds(1);
  }
  if (shift_buttons & BUTTON_SHIFTMASK_B)
    buttons |= ARCADA_BUTTONMASK_B;
  if (shift_buttons & BUTTON_SHIFTMASK_A)
    buttons |= ARCADA_BUTTONMASK_A;
  if (shift_buttons & BUTTON_SHIFTMASK_SELECT)
    buttons |= ARCADA_BUTTONMASK_SELECT;
  if (shift_buttons & BUTTON_SHIFTMASK_START)
    buttons |= ARCADA_BUTTONMASK_START;
  if (shift_buttons & BUTTON_SHIFTMASK_UP)
    buttons |= ARCADA_BUTTONMASK_UP;
  if (shift_buttons & BUTTON_SHIFTMASK_DOWN)
    buttons |= ARCADA_BUTTONMASK_DOWN;
  if (shift_buttons & BUTTON_SHIFTMASK_LEFT)
    buttons |= ARCADA_BUTTONMASK_LEFT;
  if (shift_buttons & BUTTON_SHIFTMASK_RIGHT)
    buttons |= ARCADA_BUTTONMASK_RIGHT;

#endif

#ifdef ARCADA_BUTTONPIN_START
  if (!digitalRead(ARCADA_BUTTONPIN_START)) 
    buttons |= ARCADA_BUTTONMASK_START;
#endif

#ifdef ARCADA_BUTTONPIN_SELECT
  if (!digitalRead(ARCADA_BUTTONPIN_SELECT)) 
    buttons |= ARCADA_BUTTONMASK_SELECT;
#endif

#ifdef ARCADA_BUTTONPIN_A
  if (!digitalRead(ARCADA_BUTTONPIN_A)) 
    buttons |= ARCADA_BUTTONMASK_A;
#endif

#ifdef ARCADA_BUTTONPIN_B
  if (!digitalRead(ARCADA_BUTTONPIN_B)) 
    buttons |= ARCADA_BUTTONMASK_B;
#endif

#ifdef BUTTONPIN_UP  // gpio for buttons
  if (!digitalRead(ARCADA_BUTTONPIN_UP)) 
    buttons |= ARCADA_BUTTONMASK_UP;
  if (!digitalRead(ARCADA_BUTTONPIN_DOWN))
    buttons |= ARCADA_BUTTONMASK_DOWN;
  if (!digitalRead(ARCADA_BUTTONPIN_LEFT))
    buttons |= ARCADA_BUTTONMASK_LEFT;
  if (!digitalRead(ARCADA_BUTTONPIN_RIGHT))
    buttons |= ARCADA_BUTTONMASK_RIGHT;
#else
  int16_t x = readJoystickX();
  int16_t y = readJoystickY();
  if (x > 100)  
    buttons |= ARCADA_BUTTONMASK_RIGHT;
  else if (x < -100)  
    buttons |= ARCADA_BUTTONMASK_LEFT;
  if (y > 100)  
    buttons |= ARCADA_BUTTONMASK_DOWN;
  else if (y < -100)  
    buttons |= ARCADA_BUTTONMASK_UP;
#endif

  return buttons;
}

/************************************* Filesystem stuff *******************/


/**************************************************************************/
/*!
    @brief  Initialize the filesystem, either SD or QSPI
    @return True if a filesystem exists and was found
*/
/**************************************************************************/
bool Adafruit_Arcada::filesysBegin(void) {
#if defined(USE_SD_FS)
  return FileSys.begin(SD_CS);
#elif defined(USE_QSPI_FS)
  if (!flash.begin(SPIFLASHTYPE_W25Q16BV)) {
    Serial.println("Error, failed to initialize filesys!");
    return false;
  }
  Serial.print("Flash chip JEDEC ID: 0x"); Serial.println(flash.GetJEDECID(), HEX);

  // First call begin to mount the filesystem.  Check that it returns true
  // to make sure the filesystem was mounted.
  if (!FileSys.begin()) {
    Serial.println("Failed to mount filesystem!");
    Serial.println("Was CircuitPython loaded on the board first to create the filesystem?");
    return false;
  }
  Serial.println("Mounted filesystem!");
  return true;
#else
  return false;
#endif
}

/**************************************************************************/
/*!
    @brief  Set working directory to a given path (makes file naming easier)
    @return True if was able to find a directory at that path
*/
/**************************************************************************/
bool Adafruit_Arcada::filesysCWD(const char *path) {
  if (strlen(path) >= sizeof(_cwd_path)) {    // too long!
    return false;
  }
  strcpy(_cwd_path, path);
  File dir = FileSys.open(_cwd_path);
  if (! dir) {   // couldnt open?
    return false;
  }

  if (! dir.isDirectory()) {    // not a directory or something else :(
    return false;
  }
  // ok could open and is dir
  dir.close();
  return true;
}


/**************************************************************************/
/*!
    @brief  Debugging helper, prints to Serial a list of files in a path
    @param  path A string with the filename path, must start with / e.g. "/roms".
    If nothing is passed in, we use the CWD (default is "/")
    @return -1 if was not able to open, or the number of files
*/
/**************************************************************************/
int16_t Adafruit_Arcada::filesysListFiles(const char *path) {
  if (! path) {   // use CWD!
    path = _cwd_path;
  }

  File dir = FileSys.open(path);
  char filename[SD_MAX_FILENAME_SIZE];
  int16_t num_files = 0;
    
  if (!dir) 
    return -1;

  while (1) {
    File entry =  dir.openNextFile();
    if (! entry) {
      return num_files; // no more files
    }
#if defined(USE_QSPI_FS)
    strncpy(filename, entry.name(), SD_MAX_FILENAME_SIZE);
#else
    entry.getName(filename, SD_MAX_FILENAME_SIZE);
#endif
    Serial.print(filename);
    if (entry.isDirectory()) {
      Serial.println("/");
    } else {
      num_files++;
      // files have sizes, directories do not
      Serial.print("\t\t");
      Serial.println(entry.size(), DEC);
    }
    entry.close();
  }
  return -1;
}


/**************************************************************************/
/*!
    @brief  Tests if a file exists
    @param  path A string with the filename path
    @return true or false if we can open the file
*/
/**************************************************************************/
bool Adafruit_Arcada::exists(const char *path) {
  File f = open(path);
  if (!f) return false;
  f.close();
  return true;
}

/**************************************************************************/
/*!
    @brief  Opens a file and returns the object, a wrapper for our filesystem
    @param  path A string with the filename path, must start with / e.g. "/roms"
    @return A File object, for whatever filesystem we're using
*/
/**************************************************************************/
File Adafruit_Arcada::open(const char *path, uint32_t flags) {
  if (!path) {    // Just the CWD then
    Serial.printf("Open CWD\n");
    return FileSys.open(_cwd_path, flags);
  }
  if (path[0] == '/') { // absolute path
    Serial.printf("Open absolute path %s\n", path);
    return FileSys.open(path, flags);
  }
  // otherwise, merge CWD and path
  String cwd(_cwd_path);
  String subpath(path);
  String combined = cwd + String("/") + subpath;
  char totalpath[255];
  combined.toCharArray(totalpath, 255);
  Serial.printf("Open totalpath %s\n", totalpath);
  return FileSys.open(totalpath, flags);
}


/*
uint8_t * Adafruit_Arcada::writeFileToFlash(const char *filename, uint32_t address) {
  File f = open(filename);
  if (!f) return NULL;

  uint32_t filesize = f.fileSize();
  Serial.printf("Filesize : %d bytes\n", filesize);

  uint32_t flashsize = 0;
#ifdef __SAMD51P20A__ 
  flashsize = 1024 * 1024;
#endif
  if (! flashsize) {
    Serial.println("Can't determine flash size");
    return NULL;
  }
  Serial.printf("%d bytes available\n", flashsize - address);
  
  if (8 << NVMCTRL->PARAM.bit.PSZ != FLASH_PAGE_SIZE) {
    Serial.printf("Wrong flash page size %d\n", FLASH_PAGE_SIZE);
    return NULL;
  }
  Serial.printf("Flash page size %d\n", FLASH_PAGE_SIZE);
  Serial.printf("Flash row size %d\n", FLASH_ROW_SIZE);

  uint8_t pageBuf[FLASH_ROW_SIZE];

  int i, fileremaining=filesize;

  for (i = 0; i < filesize; i += FLASH_ROW_SIZE) {

    memset(pageBuf, 0xFF, FLASH_ROW_SIZE);
    int toRead = min(fileremaining, FLASH_ROW_SIZE);
    if (f.read(pageBuf, toRead) != toRead) {
      Serial.printf("File read %d bytes failed!", toRead);
    }
    fileremaining -= toRead;

    Serial.printf("Writing %d bytes: ", toRead);
    for (int b=0; b<toRead; b++) {
      Serial.printf("0x%02X, ", pageBuf[b]);
    }
    Serial.println();
    flash_write_row((uint32_t *)(void *)i, (uint32_t *)(void *)pageBuf);

  }


  return NULL;
}


static inline wait_ready(void) {
  while (NVMCTRL->STATUS.bit.READY == 0);
}

static void flash_erase_block(uint32_t *dst) {
    wait_ready();

    // Execute "ER" Erase Row
    NVMCTRL->ADDR.reg = (uint32_t)dst;
    NVMCTRL->CTRLB.reg = NVMCTRL_CTRLB_CMDEX_KEY | NVMCTRL_CTRLB_CMD_EB;
    wait_ready();
}


#define QUAD_WORD (4 * 4)
void flash_write_words(uint32_t *dst, uint32_t *src, uint32_t n_words) {
    // Set manual page write
    NVMCTRL->CTRLA.bit.WMODE = NVMCTRL_CTRLA_WMODE_MAN;

    // Execute "PBC" Page Buffer Clear
    wait_ready();
    NVMCTRL->CTRLB.reg = NVMCTRL_CTRLB_CMDEX_KEY | NVMCTRL_CTRLB_CMD_PBC;
    wait_ready();

    while (n_words > 0) {
        // We write quad words so that we can write 256 byte blocks like UF2
        // provides. Pages are 512 bytes and would require loading data back out
        // of flash for the neighboring row.
        uint32_t len = 4 < n_words ? 4 : n_words;

        wait_ready();
        for (uint32_t i = 0; i < 4; i++) {
            if (i < len) {
                dst[i] = src[i];
            } else {
                dst[i] = 0xffffffff;
            }
        }

        // Trigger the quad word write.
        NVMCTRL->ADDR.reg = (uint32_t)dst;
        NVMCTRL->CTRLB.reg = NVMCTRL_CTRLB_CMDEX_KEY | NVMCTRL_CTRLB_CMD_WQW;

        // Advance to quad word
        dst += len;
        src += len;
        n_words -= len;
    }
}

// On the SAMD51 we can only erase 4KiB blocks of 512 byte pages. To reduce wear
// and increase flash speed we only want to erase a block at most once per
// flash. Each 256 byte row from the UF2 comes in an unknown order. So, we wait
// to erase until we see a row that varies with current memory. Before erasing,
// we cache the rows that were the same up to this point, perform the erase and
// flush the previously seen rows. Every row after will get written without
// another erase.

bool block_erased[FLASH_SIZE / NVMCTRL_BLOCK_SIZE];
bool row_same[FLASH_SIZE / NVMCTRL_BLOCK_SIZE][NVMCTRL_BLOCK_SIZE / FLASH_ROW_SIZE];

// Skip writing blocks that are identical to the existing block.
// only disable for debugging/timing
#define QUICK_FLASH 1

void flash_write_row(uint32_t *dst, uint32_t *src) {
    const uint32_t FLASH_ROW_SIZE_WORDS = FLASH_ROW_SIZE / 4;

    // The cache in Rev A isn't reliable when reading and writing to the NVM.
    NVMCTRL->CTRLA.bit.CACHEDIS0 = true;
    NVMCTRL->CTRLA.bit.CACHEDIS1 = true;

    uint32_t block = ((uint32_t) dst) / NVMCTRL_BLOCK_SIZE;
    uint8_t row = (((uint32_t) dst) % NVMCTRL_BLOCK_SIZE) / FLASH_ROW_SIZE;
#if QUICK_FLASH
    bool src_different = false;
    for (uint32_t i = 0; i < FLASH_ROW_SIZE_WORDS; ++i) {
        if (src[i] != dst[i]) {
            src_different = true;
            break;
        }
    }

    // Row is the same, quit early but keep track in case we need to erase its
    // block. This is ok after an erase because the destination will be all 1s.
    if (!src_different) {
        row_same[block][row] = true;
        return;
    }
#endif

    if (!block_erased[block]) {
        uint8_t rows_per_block = NVMCTRL_BLOCK_SIZE / FLASH_ROW_SIZE;
        uint32_t* block_address = (uint32_t *) (block * NVMCTRL_BLOCK_SIZE);

        bool some_rows_same = false;
        for (uint8_t i = 0; i < rows_per_block; i++) {
            some_rows_same = some_rows_same || row_same[block][i];
        }
        uint32_t row_cache[rows_per_block][FLASH_ROW_SIZE_WORDS];
        if (some_rows_same) {
            for (uint8_t i = 0; i < rows_per_block; i++) {
                if(row_same[block][i]) {
                    memcpy(row_cache[i], block_address + i * FLASH_ROW_SIZE_WORDS, FLASH_ROW_SIZE);
                }
            }
        }
        flash_erase_block(dst);
        block_erased[block] = true;
        if (some_rows_same) {
            for (uint8_t i = 0; i < rows_per_block; i++) {
                if(row_same[block][i]) {
                    // dst is a uint32_t pointer so we add the number of words,
                    // not bytes.
                    flash_write_words(block_address + i * FLASH_ROW_SIZE_WORDS, row_cache[i], FLASH_ROW_SIZE_WORDS);
                }
            }
        }
    }

    flash_write_words(dst, src, FLASH_ROW_SIZE_WORDS);

    // Don't return until we're done writing in case something after us causes
    // a reset.
    wait_ready();
}

*/

/**************************************************************************/
/*!
    @brief  Read the light sensor onboard if there is one
    @return 0 (darkest) to 1023 (brightest) or 0 if there is no sensor
*/
/**************************************************************************/
uint16_t readLightSensor(void) {
#if defined(LIGHT_SENSOR)
  return analogRead(LIGHT_SENSOR);
#else
  return 0;
#endif
}


void Adafruit_Arcada::fillScreen(uint16_t color) {
  return tft.fillScreen(color);
}

void Adafruit_Arcada::invertDisplay(bool flag) {
  return tft.invertDisplay(flag);
}

bool Adafruit_Arcada::createFrameBuffer(uint16_t width, uint16_t height) {
  _framebuffer = (uint16_t *)malloc(width * height * 2);
  if (!_framebuffer) return false;
  _framebuf_width = width;
  _framebuf_height = height;
  return true;
}

uint16_t * Adafruit_Arcada::getFrameBuffer(void) {
  return _framebuffer;
}

bool Adafruit_Arcada::blitFrameBuffer(uint16_t x, uint16_t y, bool blocking) {
  if (!_framebuffer) return false;

  if (! _first_frame) {
    tft.endWrite(); // End transaction from any prior callback
    _first_frame = false;
  }

  tft.startWrite(); // Start new display transaction
  tft.setAddrWindow(x, y, _framebuf_width, _framebuf_height);
  tft.writePixels(_framebuffer, _framebuf_width*_framebuf_height, blocking); // immediate return;
  return true;
}

					
